/* gvSIG. Sistema de Informacin Geogrfica de la Generalitat Valenciana
 *
 * Copyright (C) 2006 IVER T.I. and Generalitat Valenciana.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,USA.
 */
package com.iver.cit.gvsig.fmap.layers;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics2D;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.awt.image.BufferedImage;
import java.util.ArrayList;
import java.util.Iterator;

import org.gvsig.georeferencing.utils.GeoPointPersistence;
import org.gvsig.georeferencing.utils.PointManager;

import com.iver.andami.PluginServices;
import com.iver.cit.gvsig.fmap.DriverException;
import com.iver.cit.gvsig.fmap.ViewPort;
import com.iver.cit.gvsig.fmap.crs.CRSFactory;
import com.iver.cit.gvsig.project.documents.view.gui.View;
import com.iver.utiles.XMLEntity;
import com.iver.utiles.swing.threads.Cancellable;


/**
 * Clase de capa de marcado de puntos sobre una vista. Dibuja un puntero sobre
 * cada punto en la vista. Esta capa hereda de GraphicLayer que es una capa que est
 * metida de forma fija en FMap. Con esto conseguimos que nuestra capa de puntos
 * no aparezca en el TOC y siempre se dibuje por encima de todas las capas. Tiene
 * el inconveniente de que est en FMap y si hacemos un dibujado de las capas
 * sobre un panel que no sea la vista de gvSIG este GraphicLayer no aparece.
 * Si queremos hacer este tipo de dibujados tendremos que dibujar todas las capas
 * con el draw de FLayers y al final llamar al draw de FLyrPoints.
 *
 * @author Nacho Brodin (brodin_ign@gva.es)
 */
public class FLyrPoints extends GraphicLayer {

	//**********************Params********************************
	/**
	 * Diametro del centro de la cruz
	 */
	private final int 				DIAM_CIRCLE = 18;
	/**
	 * ltima herramienta seleccionada. Despus de marcar el segundo punto en la vista ser la
	 * herramienta por defecto.
	 */
	private String 					lastTool = "zoomIn";
	//**********************End Params****************************

	//**********************Vars**********************************
	private ArrayList 				pointList = new ArrayList();
	private boolean 				showNumber = true;
	private FLyrPointsState 		state = null;
	private IPersistence 			geoPointPersistence = null;
	private FLyrGeoRaster 			lyrGeoRaster = null;
	public static boolean			isDrawing = false;
	private	PointManager			pointManager = null;
	private	boolean					lyrVisible = true;
	private	boolean					lyrHideViewOnly = true;
	private Point2D[]				transformPointList = null;
	private boolean					withoutBlue = false;
	private boolean					withoutRed = false;
	//**********************End Vars**********************************

	//**********************Classes***********************************
	/**
	 * Estado de la capa de puntos.
	 * @author Nacho Brodin (brodin_ign@gva.es)
	 */
	public class FLyrPointsState{
		public ArrayList pointList = new ArrayList();
		public boolean showNumber = true;
		public String lastTool = "zoomIn";
	}
	//**********************End Classes*******************************

	//**********************Methods***********************************
	/**
	 * Constructor
	 * @param flyGeoRaster
	 */
	public FLyrPoints(FLyrGeoRaster lyGeoRaster, IPersistence persistence){
		this.lyrGeoRaster = lyGeoRaster;
		geoPointPersistence = persistence;
	}

	/**
	 * Salva el estado actual de la capa de puntos. Podr ser recuperado
	 * este ltimo estado salvado con la funcin recoveryState
	 *
	 */
	public void saveState(){
		if(state == null)
			state = new FLyrPointsState();
		else
			state.pointList.clear();

		for(int i=0;i<this.length();i++)
			state.pointList.add(((GeoPoint)pointList.get(i)).cloneGeoPoint());

		state.showNumber = showNumber;
		state.lastTool = lastTool;
	}

	/**
	 * Recupera el estado de la capa de puntos desde la ltima vez que se
	 * ejecut la funcin saveState. Si no hay ningn estdo salvado es que
	 * al principio estaba la capa vacia por lo que reiniciamos la capa de puntos.
	 */
	public void recoveryState(){
		if(state != null){
			pointList.clear();
			for(int i=0;i<state.pointList.size();i++)
				pointList.add(((GeoPoint)state.pointList.get(i)).cloneGeoPoint());

			showNumber = state.showNumber;
			lastTool = state.lastTool;
		}
	}

	/**
	 * Asigna el array de puntos
	 * @param list
	 */
	public void setListPoint(ArrayList list){
		pointList = list;
	}

	/**
	 * Dibujado de la capa de raster georeferenciado aplicando la
	 * transformacin del viewPort.
	 */
	public void draw(BufferedImage image, Graphics2D g, ViewPort vp,
			Cancellable cancel,double scale) throws DriverException {

		Point2D pto = null;

		if(lyrHideViewOnly && lyrVisible && lyrGeoRaster != null){
			//vp = lyrGeoRaster.toImageCoord(vp);
			synchronized (this) {
				for(int i=0; i<pointList.size();i++){
					//Punto de la imagen
					if(((GeoPoint)pointList.get(i)).active){
							pto = ((GeoPoint)pointList.get(i)).pixelPoint;
							if(pto != null){
								Point2D p = lyrGeoRaster.img2World(pto);
								p = vp.fromMapPoint(p);
								drawPixelCrux(g, p, i);
							}

						//Punto de la vista
							pto = ((GeoPoint)pointList.get(i)).mapPoint;
							if(pto != null){
								Point2D p = null;
								p = vp.fromMapPoint(pto);
								drawMapCrux(g, p, i);
							}
					}
				}
			}
		}
	}

	/**
	 * Dibuja el grfico que corresponde a los puntos de coordenadas relativas a la imagen de origen
	 * @param g Graphics sobre el que se dibuja
	 * @param vp	Viewport
	 * @param cancel
	 */
	public void drawPixelCrux(Graphics2D g, ViewPort vp, Cancellable cancel){
		Point2D pto = null;
		if(lyrVisible && lyrGeoRaster != null){
			for(int i=0; i<pointList.size();i++){
				if(((GeoPoint)pointList.get(i)).active){
					pto = ((GeoPoint)pointList.get(i)).pixelPoint;
					if(pto != null){
						Point2D p = lyrGeoRaster.img2World(pto);
						p = vp.fromMapPoint(p);
						drawPixelCrux(g, p, i);
					}
				}
			}
		}
	}

	/**
	 * Dibuja el grfico que corresponde a los puntos de coordenadas de mundo real.
	 * @param g Graphics sobre el que se dibuja
	 * @param vp	Viewport
	 * @param cancel
	 */
	public void drawMapCrux(Graphics2D g, ViewPort vp, Cancellable cancel){
		Point2D pto = null;
		if(lyrVisible && lyrGeoRaster != null){
			for(int i=0; i<pointList.size();i++){
				if(((GeoPoint)pointList.get(i)).active){
					pto = ((GeoPoint)pointList.get(i)).mapPoint;
					if(pto != null){
						Point2D p = null;
						p = vp.fromMapPoint(pto);
						drawMapCrux(g, p, i);
					}
				}
			}
		}
	}

	/**
	 * Dibuja sobre el graphics pasado la cruz del punto que marca
	 * el pixel de la imagen.
	 * @param g Graphics
	 */
	private void drawPixelCrux(Graphics2D g, Point2D p, int pointNumber){
		int dpto = (DIAM_CIRCLE >> 1);
		int incr = 5;
		g.setColor(Color.WHITE);
		g.drawOval(	(int)p.getX() - dpto + 1,
					(int)p.getY() - dpto + 1,
					DIAM_CIRCLE - 2,
					DIAM_CIRCLE - 2);
		g.drawLine((int)p.getX() - incr, (int)p.getY() - 1, (int)p.getX() - 1, (int)p.getY() - 1);
		g.drawLine((int)p.getX() - incr, (int)p.getY() + 1, (int)p.getX() - 1, (int)p.getY() + 1);

		g.drawLine((int)p.getX() + incr, (int)p.getY() - 1, (int)p.getX() + 1, (int)p.getY() - 1);
		g.drawLine((int)p.getX() + incr, (int)p.getY() + 1, (int)p.getX() + 1, (int)p.getY() + 1);

		g.drawLine((int)p.getX() - 1, (int)p.getY() - incr, (int)p.getX() - 1, (int)p.getY() - 1);
		g.drawLine((int)p.getX() + 1, (int)p.getY() - incr, (int)p.getX() + 1, (int)p.getY() - 1);

		g.drawLine((int)p.getX() - 1, (int)p.getY() + incr, (int)p.getX() - 1, (int)p.getY() + 1);
		g.drawLine((int)p.getX() + 1, (int)p.getY() + incr, (int)p.getX() + 1, (int)p.getY() + 1);

		g.setColor(Color.red);
		g.drawOval(	(int)p.getX() - dpto,
					(int)p.getY() - dpto,
					DIAM_CIRCLE,
					DIAM_CIRCLE);
		g.drawLine((int)p.getX(), (int)p.getY() - dpto - incr, (int)p.getX(), (int)p.getY() + dpto + incr);
		g.drawLine((int)p.getX() - dpto - incr, (int)p.getY(), (int)p.getX() + dpto + incr, (int)p.getY());

		if(showNumber){
			String pt = String.valueOf(pointNumber + 1);
			int ptX = (int)(p.getX() + dpto + 1);
			int ptY = (int)(p.getY() + dpto - 1);
			g.setColor(Color.WHITE);
			for (int i= -1; i<2; i++)
				for (int j= -1; j<2; j++)
					g.drawString(pt, ptX + i, ptY + j );
			g.setColor(Color.red);
			g.drawString(pt, ptX, ptY );
		}
	}

	/**
	 * Dibuja sobre el graphics pasado la cruz del punto que marca
	 * la coordenada de la vista.
	 * @param g Graphics
	 */
	private void drawMapCrux(Graphics2D g, Point2D p, int pointNumber){
		int dpto = (DIAM_CIRCLE >> 1);
		int incr = 5;
		g.setColor(Color.WHITE);
		g.drawRect(	(int)p.getX() - dpto + 1,
					(int)p.getY() - dpto + 1,
					DIAM_CIRCLE - 2,
					DIAM_CIRCLE - 2);
		g.drawLine((int)p.getX() - incr, (int)p.getY() - 1, (int)p.getX() - 1, (int)p.getY() - 1);
		g.drawLine((int)p.getX() - incr, (int)p.getY() + 1, (int)p.getX() - 1, (int)p.getY() + 1);

		g.drawLine((int)p.getX() + incr, (int)p.getY() - 1, (int)p.getX() + 1, (int)p.getY() - 1);
		g.drawLine((int)p.getX() + incr, (int)p.getY() + 1, (int)p.getX() + 1, (int)p.getY() + 1);

		g.drawLine((int)p.getX() - 1, (int)p.getY() - incr, (int)p.getX() - 1, (int)p.getY() - 1);
		g.drawLine((int)p.getX() + 1, (int)p.getY() - incr, (int)p.getX() + 1, (int)p.getY() - 1);

		g.drawLine((int)p.getX() - 1, (int)p.getY() + incr, (int)p.getX() - 1, (int)p.getY() + 1);
		g.drawLine((int)p.getX() + 1, (int)p.getY() + incr, (int)p.getX() + 1, (int)p.getY() + 1);

		g.setColor(new Color(45, 8 , 165));
		g.drawRect(	(int)p.getX() - dpto,
					(int)p.getY() - dpto,
					DIAM_CIRCLE,
					DIAM_CIRCLE);
		g.drawLine((int)p.getX(), (int)p.getY() - dpto - incr, (int)p.getX(), (int)p.getY() + dpto + incr);
		g.drawLine((int)p.getX() - dpto - incr, (int)p.getY(), (int)p.getX() + dpto + incr, (int)p.getY());
		if(showNumber){
			String pt = String.valueOf(pointNumber + 1);
			int ptX = (int)(p.getX() + dpto + 1);
			int ptY = (int)(p.getY() - dpto - 1);
			g.setColor(Color.WHITE);
			for (int i= -1; i<2; i++)
				for (int j= -1; j<2; j++)
					g.drawString(pt, ptX + i, ptY + j );
			g.setColor(new Color(45, 8 , 165));
			g.drawString(String.valueOf(pointNumber + 1), (int)(p.getX() + dpto + 1), (int)(p.getY() - dpto - 1) );
		}
	}

	/* (non-Javadoc)
	 * @see com.iver.cit.gvsig.fmap.layers.FLayer#print(java.awt.Graphics2D, com.iver.cit.gvsig.fmap.ViewPort, com.iver.cit.gvsig.fmap.operations.Cancellable, double)
	 */
	public void print(Graphics2D g, ViewPort viewPort, Cancellable cancel,
			double scale) throws DriverException {

	}

	/**
	 * Elimina los puntos de la lista que no tiene las dos coordenadas asignadas.
	 */
	public void clean(){
		Iterator iter = pointList.iterator();
		while (iter.hasNext()) {
			GeoPoint gp = (GeoPoint) iter.next();
			if(gp.mapPoint == null || gp.pixelPoint == null)
				iter.remove();
		}
	}
	/**
	 * Aade un punto a la lista
	 * @param point punto para la lista
	 */
	public GeoPoint addPoint(Point2D pixel, Point2D map){
		GeoPoint p = new GeoPoint(pixel, map);
		pointList.add(p);
		return p;
	}

	/**
	 * Obtiene el punto de la posicin pos
	 */
	public GeoPoint getPoint(int pos){
		try{
			return (GeoPoint)pointList.get(pos);
		}catch(IndexOutOfBoundsException exc){
			return null;
		}
	}

	/**
	 * Elimina la lista de puntos almacenada
	 */
	public void clear(){
		pointList = new ArrayList();
	}

	/**
	 * Elimina el punto de la posicin indicada por el parmetro pos
	 * @param pos	Posicin del punto a eliminar
	 */
	public void remove(int pos){
		pointList.remove(pos);
	}

	/**
	 *Elimina el ltimo punto de la lista.
	 */
	public void delLastPoint(){
		pointList.remove(pointList.size() - 1);
	}

	/**
	 * Funcin que obtiene si un punto es nuevo o no. Un punto es nuevo si todavia
	 * no han sido asignadas todas sus coordenadas.
	 * @param n Nmero de punto a obtener si es nuevo
	 * @return	true si es nuevo o false si no lo es
	 */
	public boolean isNewPoint(int n){
		if(this.getPoint(n) == null)
			return false;
		if(this.getPoint(n).pixelPoint == null || this.getPoint(n).mapPoint == null)
			return true;
		return false;
	}

	/**
	 * Devuelve el nmero de puntos de la capa
	 * @return entero que representa el nmero de puntos de la capa
	 */
	public int length(){
		return pointList.size();
	}

	/**
	 * Actualiza un punto de la lista de una posicin determinada
	 * @param point punto para la lista
	 */
	public void updatePoint(Point2D pixel, Point2D map, int pos){
		GeoPoint gp = (GeoPoint)pointList.get(pos);
		if(pixel != null)
			gp.pixelPoint = pixel;
		if(map != null)
			gp.mapPoint = map;
	}

	/**
	 * Calcula el RMS para el punto de la posicin pos
	 * @param pos	Posicin
	 * @return	valor RMS para el punto solicitado
	 */
	public double calcDoubleRMS(int pos){
		if(getPoint(pos).active && transformPointList != null && getPoint(pos).mapPoint != null){
			try{
				double a = (transformPointList[pos].getX() - getPoint(pos).mapPoint.getX());
				double b = (transformPointList[pos].getY() - getPoint(pos).mapPoint.getY());
				return Math.sqrt(a*a + b*b);
			}catch(ArrayIndexOutOfBoundsException e){
				//No hacemos nada pq ha intentado obtener un punto que no existe. Hace un return 0
			}catch(NullPointerException e){
				//No hacemos nada pq el punto an no est completo y est intentando obtener un valor que no existe.Hace un return 0
			}
		}
		return -0D;
	}

	/**
	 * Calcula el RMS para el punto de la posicin pos
	 * @param pos	Posicin
	 * @return	RMS en forma de String
	 */
	public String calcRMS(int pos){
		return String.valueOf(calcDoubleRMS(pos));
	}

	/**
	 * Calcula el residuo en X para el punto de la posicin pos
	 * @param pos	Posicin
	 * @return		residuo Y en forma de double
	 */
	public double calcDoubleResX(int pos){
		if(	getPoint(pos).active &&
			transformPointList != null &&
			transformPointList[pos] != null &&
			getPoint(pos).mapPoint != null){
			try{
				double a = (transformPointList[pos].getX() - getPoint(pos).mapPoint.getX());
				return Math.sqrt(a*a);
			}catch(ArrayIndexOutOfBoundsException e){
				//No hacemos nada pq ha intentado obtener un punto que no existe. Hace un return 0
			}catch(NullPointerException e){
				//No hacemos nada pq el punto an no est completo y est intentando obtener un valor que no existe.Hace un return 0
			}
		}
		return -0;
	}

	/**
	 * Calcula el residuo en X para el punto de la posicin pos
	 * @param pos	Posicin
	 * @return	residuo Y en forma de String
	 */
	public String calcResX(int pos){
		return String.valueOf(calcDoubleResX(pos));
	}

	/**
	 * Calcula el residuo en Y para el punto de la posicin pos
	 * @param pos	Posicin
	 * @return	residuo Y en forma de double
	 */
	public double calcDoubleResY(int pos){
		if(	getPoint(pos).active &&
			transformPointList != null &&
			transformPointList[pos] != null &&
			getPoint(pos).mapPoint != null){
			try{
				double b = (transformPointList[pos].getY() - getPoint(pos).mapPoint.getY());
				return Math.sqrt(b*b);
			}catch(ArrayIndexOutOfBoundsException e){
				//No hacemos nada pq ha intentado obtener un punto que no existe. Hace un return 0
			}catch(NullPointerException e){
				//No hacemos nada pq el punto an no est completo y est intentando obtener un valor que no existe.Hace un return 0
			}
		}
		return -0D;
	}

	/**
	 * Calcula el residuo en Y para el punto de la posicin pos
	 * @param pos	Posicin
	 * @return	residuo Y en forma de String
	 */
	public String calcResY(int pos){
		return String.valueOf(calcDoubleResY(pos));
	}

	/**
	 * Calcula el error RMS para todos los puntos
	 * @return	Error RMS total en forma de String
	 */
	public double calcDoubleTotal(){
		double add = 0D;
		int len = 0;
		try{
			for(int i=0;i<transformPointList.length;i++){
				add += calcDoubleRMS(i);
				if(getPoint(i).active)
					len ++;
			}
		}catch(NullPointerException e){
			//transformPointList es null por lo que no hay puntos a transformar devolviendo 0.
			return 0D;
		}
		return Math.sqrt(add/len);
	}

	/**
	 * Calcula el error RMS para todos los puntos
	 * @return	Error RMS total en forma de String
	 */
	public String calcTotal(){
		return String.valueOf(calcDoubleTotal());
	}

	/**
	 * Muestra en consola los puntos de la capa
	 */
	public void showPoints(){
		for(int i=0;i<pointList.size();i++)
			((GeoPoint)pointList.get(i)).show();
	}

	/**
	 * Salva la lista de puntos sobre el fichero que se la pasa por
	 * parmetro
	 * @param file Nombre del fichero
	 */
	public void PointList2XML(String file){;
		geoPointPersistence.savePointList(file);
	}

	/**
	 * Recupera una lista de puntos desde un fichero XML
	 * @param file Nombre del fichero
	 */
	public void XML2PointList(String file){
		geoPointPersistence.loadPointList(file);
	}

	/**
	 * Salva la lista de puntos en forma tabular sobre el fichero que se la pasa por
	 * parmetro
	 * @param file Nombre del fichero
	 */
	public void PointList2CSV(String file){;
		geoPointPersistence.saveCSVPointList(file);
	}

	/**
	 * Recupera la lista de puntos desde un fichero CSV
	 * @param file Nombre del fichero
	 */
	public void PointListFromCSV(String file){;
		geoPointPersistence.loadCSVPointList(file);
	}

	/* (non-Javadoc)
	 * @see com.iver.cit.gvsig.fmap.layers.StatusRasterInterface#setXMLEntity(com.iver.utiles.XMLEntity)
	 */
	public void setXMLEntity(XMLEntity xml)throws XMLException {
		int nPoints = -1;
		if (xml.contains("raster.geoPoints.nPoints")) {
			nPoints = xml.getIntProperty("raster.geoPoints.nPoints");
			GeoPoint point = null;
			Point2D pixel = new Point2D.Double(), map = new Point2D.Double();
			if(nPoints > 0){

				for(int i = 0;i<nPoints;i++){

					if (xml.contains("raster.geoPoint."+i+".pixelX") && xml.contains("raster.geoPoint."+i+".pixelY"))
						pixel.setLocation(xml.getDoubleProperty("raster.geoPoint."+i+".pixelX"), xml.getDoubleProperty("raster.geoPoint."+i+".pixelY"));

					if (xml.contains("raster.geoPoint."+i+".mapX") && xml.contains("raster.geoPoint."+i+".mapY"))
						map.setLocation(xml.getDoubleProperty("raster.geoPoint."+i+".mapX"), xml.getDoubleProperty("raster.geoPoint."+i+".mapY"));

					point = this.addPoint(pixel, map);

					if (xml.contains("raster.geoPoint."+i+".active"))
						point.active = xml.getBooleanProperty("raster.geoPoint."+i+".active");

					point.leftCenterPoint = new Point2D.Double();
					if (xml.contains("raster.geoPoint."+i+".leftCenterX") && xml.contains("raster.geoPoint."+i+".leftCenterY"))
						point.leftCenterPoint.setLocation(xml.getDoubleProperty("raster.geoPoint."+i+".leftCenterX"), xml.getDoubleProperty("raster.geoPoint."+i+".leftCenterY"));

					point.rightCenterPoint = new Point2D.Double();
					if (xml.contains("raster.geoPoint."+i+".rightCenterX") && xml.contains("raster.geoPoint."+i+".rightCenterY"))
						point.rightCenterPoint.setLocation(xml.getDoubleProperty("raster.geoPoint."+i+".rightCenterX"), xml.getDoubleProperty("raster.geoPoint."+i+".rightCenterY"));

					point.leftViewPort = new ViewPort(null);
					point.rightViewPort = new ViewPort(null);

					double x = 0D, y = 0D, w = 0D, h = 0D;
					int width = 0, height = 0;
					Rectangle2D r = new Rectangle2D.Double();
					if (xml.contains("raster.geoPoint."+i+".leftViewPort.proj"))
						point.leftViewPort.setProjection(CRSFactory.getCRS(xml.getStringProperty("raster.geoPoint."+i+".leftViewPort.proj")));
					if (xml.contains("raster.geoPoint."+i+".leftViewPort.x"))
						x = xml.getDoubleProperty("raster.geoPoint."+i+".leftViewPort.x");
					if (xml.contains("raster.geoPoint."+i+".leftViewPort.y"))
						y = xml.getDoubleProperty("raster.geoPoint."+i+".leftViewPort.y");
					if (xml.contains("raster.geoPoint."+i+".leftViewPort.w"))
						w = xml.getDoubleProperty("raster.geoPoint."+i+".leftViewPort.w");
					if (xml.contains("raster.geoPoint."+i+".leftViewPort.h"))
						h = xml.getDoubleProperty("raster.geoPoint."+i+".leftViewPort.h");
					r.setRect(x, y, w, h);
					point.leftViewPort.setExtent(r);
					if (xml.contains("raster.geoPoint."+i+".leftViewPort.imgWidth"))
						width = xml.getIntProperty("raster.geoPoint."+i+".leftViewPort.imgWidth");
					if (xml.contains("raster.geoPoint."+i+".leftViewPort.imgHeight"))
						height = xml.getIntProperty("raster.geoPoint."+i+".leftViewPort.imgHeight");
					point.leftViewPort.setImageSize(new Dimension(width, height));

					r = new Rectangle2D.Double();
					if (xml.contains("raster.geoPoint."+i+".rightViewPort.proj"))
						point.rightViewPort.setProjection(CRSFactory.getCRS(xml.getStringProperty("raster.geoPoint."+i+".rightViewPort.proj")));
					if (xml.contains("raster.geoPoint."+i+".rightViewPort.x"))
						x = xml.getDoubleProperty("raster.geoPoint."+i+".rightViewPort.x");
					if (xml.contains("raster.geoPoint."+i+".rightViewPort.y"))
						y = xml.getDoubleProperty("raster.geoPoint."+i+".rightViewPort.y");
					if (xml.contains("raster.geoPoint."+i+".rightViewPort.w"))
						w = xml.getDoubleProperty("raster.geoPoint."+i+".rightViewPort.w");
					if (xml.contains("raster.geoPoint."+i+".rightViewPort.h"))
						h = xml.getDoubleProperty("raster.geoPoint."+i+".rightViewPort.h");
					r.setRect(x, y, w, h);
					point.rightViewPort.setExtent(r);
					if (xml.contains("raster.geoPoint."+i+".rightViewPort.imgWidth"))
						width = xml.getIntProperty("raster.geoPoint."+i+".rightViewPort.imgWidth");
					if (xml.contains("raster.geoPoint."+i+".rightViewPort.imgHeight"))
						height = xml.getIntProperty("raster.geoPoint."+i+".rightViewPort.imgHeight");
					point.rightViewPort.setImageSize(new Dimension(width, height));

				}
			}
		}
	}

	/* (non-Javadoc)
	 * @see com.iver.cit.gvsig.fmap.layers.StatusRasterInterface#getXMLEntity(com.iver.utiles.XMLEntity)
	 */
	public void getXMLEntity(XMLEntity xml) throws XMLException {
		xml.putProperty("raster.geoPoints.nPoints", getCountPoints());
		for(int i=0;i<this.getCountPoints();i++){
			xml.putProperty("raster.geoPoint."+i+".pixelX", getPoint(i).pixelPoint.getX());
			xml.putProperty("raster.geoPoint."+i+".pixelY", getPoint(i).pixelPoint.getY());
			xml.putProperty("raster.geoPoint."+i+".mapX", getPoint(i).mapPoint.getX());
			xml.putProperty("raster.geoPoint."+i+".mapY", getPoint(i).mapPoint.getX());

			xml.putProperty("raster.geoPoint."+i+".active", getPoint(i).active);

			xml.putProperty("raster.geoPoint."+i+".leftCenterX", getPoint(i).leftCenterPoint.getX());
			xml.putProperty("raster.geoPoint."+i+".leftCenterY", getPoint(i).leftCenterPoint.getY());
			xml.putProperty("raster.geoPoint."+i+".rightCenterX", getPoint(i).rightCenterPoint.getX());
			xml.putProperty("raster.geoPoint."+i+".rightCenterY", getPoint(i).rightCenterPoint.getY());

			xml.putProperty("raster.geoPoint."+i+".leftViewPort.proj", getPoint(i).leftViewPort.getProjection().getAbrev());
			xml.putProperty("raster.geoPoint."+i+".leftViewPort.x", getPoint(i).leftViewPort.getExtent().getX());
			xml.putProperty("raster.geoPoint."+i+".leftViewPort.y", getPoint(i).leftViewPort.getExtent().getY());
			xml.putProperty("raster.geoPoint."+i+".leftViewPort.w", getPoint(i).leftViewPort.getExtent().getWidth());
			xml.putProperty("raster.geoPoint."+i+".leftViewPort.h", getPoint(i).leftViewPort.getExtent().getHeight());
			xml.putProperty("raster.geoPoint."+i+".leftViewPort.imgWidth", getPoint(i).leftViewPort.getImageWidth());
			xml.putProperty("raster.geoPoint."+i+".leftViewPort.imgHeight", getPoint(i).leftViewPort.getImageHeight());

			xml.putProperty("raster.geoPoint."+i+".rightViewPort.proj", getPoint(i).rightViewPort.getProjection().getAbrev());
			xml.putProperty("raster.geoPoint."+i+".rightViewPort.x", getPoint(i).rightViewPort.getExtent().getX());
			xml.putProperty("raster.geoPoint."+i+".rightViewPort.y", getPoint(i).rightViewPort.getExtent().getY());
			xml.putProperty("raster.geoPoint."+i+".rightViewPort.w", getPoint(i).rightViewPort.getExtent().getWidth());
			xml.putProperty("raster.geoPoint."+i+".rightViewPort.h", getPoint(i).rightViewPort.getExtent().getHeight());
			xml.putProperty("raster.geoPoint."+i+".rightViewPort.imgWidth", getPoint(i).rightViewPort.getImageWidth());
			xml.putProperty("raster.geoPoint."+i+".rightViewPort.imgHeight", getPoint(i).rightViewPort.getImageHeight());
		}
	}
	//**********************End Methods********************************

	//**********************Setters & Getters**************************
	/**
	 * Asigna el viewPort de la mini imagen de la derecha en
	 * el punto de la posicin que se le pasa por parmetro
	 * @param pos Posicin del punto
	 * @param vp ViewPort asignado
	 */
	public void setRightViewPort(int pos, ViewPort vp){
		getPoint(pos).rightViewPort = vp;
	}
	/**
	 * @return Returns the showNumber.
	 */
	public boolean isShowNumber() {
		return showNumber;
	}

	/**
	 * @param showNumber The showNumber to set.
	 */
	public void setShowNumber(boolean showNumber) {
		this.showNumber = showNumber;
	}

	/**
	 * Asigna el punto central de la mini imagen de la izquierda
	 * en el punto de la posicin que se le pasa por parmetro
	 * @param pos Posicin del punto
	 * @param p Punto asignado
	 */
	public void setLeftCenterPoint(int pos, Point2D p){
		try{
			getPoint(pos).leftCenterPoint = p;
		}catch(NullPointerException ex){
			//Si getPoint devuelve null es que hemos intentado acceder a una posicin que no
			//existe por lo que no asignamos ningun valor.
		}
	}

	/**
	 * Asigna el punto central de la mini imagen de la derecha
	 * en el punto de la posicin que se le pasa por parmetro
	 * @param pos Posicin del punto
	 * @param p Punto asignado
	 */
	public void setRightCenterPoint(int pos, Point2D p){
		try{
			getPoint(pos).rightCenterPoint = p;
		}catch(NullPointerException ex){
			//Si getPoint devuelve null es que hemos intentado acceder a una posicin que no
			//existe por lo que no asignamos ningun valor.
		}
	}

	/**
	 * Asigna el viewPort de la mini imagen de la izquierda en
	 * el punto de la posicin que se le pasa por parmetro
	 * @param pos Posicin del punto
	 * @param vp ViewPort asignado
	 */
	public void setLeftViewPort(int pos, ViewPort vp){
		try{
			getPoint(pos).leftViewPort = vp;
		}catch(NullPointerException ex){
			//Si getPoint devuelve null es que hemos intentado acceder a una posicin que no
			//existe por lo que no asignamos ningun valor.
		}
	}

	/**
	 * Asigna el zoom al punto de una posicin. Este zoom es el que corresponde
	 * a la miniimagen de la izquierda, es decir la que tiene las coordenadas en pixeles.
	 * @param pos Posicin del punto a asignar el zoom
	 * @param zoom Zoom a asignar.
	 */
	public void setZoomLeft(int pos, double zoom){
		try{
			getPoint(pos).zoomLeft = zoom;
		}catch(NullPointerException ex){
			//Si getPoint devuelve null es que hemos intentado acceder a una posicin que no
			//existe por lo que no asignamos ningun valor.
		}
	}

	/**
	 * Asigna el zoom al punto de una posicin. Este zoom es el que corresponde
	 * a la miniimagen de la derecha, es decir la que tiene las coordenadas relativas
	 * al mundo real.
	 * @param pos Posicin del punto a asignar el zoom
	 * @param zoom Zoom a asignar.
	 */
	public void setZoomRight(int pos, double zoom){
		try{
			getPoint(pos).zoomRight = zoom;
		}catch(NullPointerException ex){
			//Si getPoint devuelve null es que hemos intentado acceder a una posicin que no
			//existe por lo que no asignamos ningun valor.
		}
	}

	/**
	 * Obtiene el zoom de un punto y posicin. Este zoom es el que corresponde
	 * a la miniimagen de la izquierda, es decir la que tiene las coordenadas en pixeles.
	 * @param pos Posicin del punto a asignar el zoom
	 * @return Zoom a asignado.
	 */
	public double getZoomLeft(int pos){
		try{
			return getPoint(pos).zoomLeft;
		}catch(NullPointerException ex){
			//Si getPoint devuelve null es que hemos intentado acceder a una posicin que no
			//existe por lo que devolvemos un zoom de 1.0.
			return 1.0;
		}
	}

	/**
	 * Obtiene el zoom de un punto y posicin. Este zoom es el que corresponde
	 * a la miniimagen de la derecha, es decir la que tiene las coordenadas relativas
	 * al mundo real.
	 * @param pos Posicin del punto a asignar el zoom
	 * @return Zoom a asignado.
	 */
	public double getZoomRight(int pos){
		try{
			return getPoint(pos).zoomRight;
		}catch(NullPointerException ex){
			//Si getPoint devuelve null es que hemos intentado acceder a una posicin que no
			//existe por lo que devolvemos un zoom de 1.0.
			return 1.0;
		}
	}

	/**
	 * Devuelve el nmero de puntos de la lista
	 * @return entero que representa el nmero de puntos
	 */
	public int getCountPoints(){
		if(pointList != null)
			return pointList.size();
		else
			return 0;
	}

	/**
	 * Devuelve el nmero de puntos de la lista que estn activos
	 * @return entero que representa el nmero de puntos
	 */
	public int getCountActivePoints(){
		int activePoints = 0;
		if(pointList != null){
			for(int i = 0; i< pointList.size(); i++)
				if(getPoint(i).active)
					activePoints ++;
			return activePoints;
		}else
			return 0;
	}

	/**
	 * Activa o desactiva el punto de una posicin de la lista
	 * @param n	Nmero de punto
	 * @param active	true activo, false desactivo
	 */
	public void setPointActive(int n, boolean active){
		if(n < 0 || n >= this.length())
			return;
		((GeoPoint)pointList.get(n)).active = active;
	}

	/**
	 * Dice si un punto de la lista est activo o no
	 * @param n	Posicin del punto
	 * @return	true si est activo y false si no lo est.
	 */
	public boolean isPointActive(int n){
		if(n < 0 || n >= this.length())
			return false;
		return ((GeoPoint)pointList.get(n)).active;
	}

	/**
	 * Obtiene el extent de la capa
	 * @return extent en Rectangle2D.
	 */
	public Rectangle2D getFullExtent()throws DriverException {
		View theView = (View) PluginServices.getMDIManager().getActiveWindow();
		return theView.getMapControl().getMapContext().getViewPort().getExtent();
	}

	/**
	 * @return Returns the lastTool.
	 */
	public String getLastTool() {
		return lastTool;
	}

	/**
	 * @param lastTool The lastTool to set.
	 */
	public void setLastTool(String lastTool) {
		this.lastTool = lastTool;
	}

	/**
	 *Asigna el extent para las mini imagenes. Estas deben ser recuperadas cuando se selecciona
	 *un punto
	 */
	public void setMiniExtent(int nPoint, Point2D centerPoint, ViewPort vp, boolean isRight){
		ViewPort viewPortBlock = new ViewPort(vp.getProjection());
		viewPortBlock.setExtent(vp.getExtent());
		viewPortBlock.setImageSize(vp.getImageSize());
		viewPortBlock.refreshExtent();

		if(isRight){
			((GeoPoint)pointList.get(nPoint)).rightCenterPoint = (Point2D)centerPoint.clone();
			((GeoPoint)pointList.get(nPoint)).rightViewPort = viewPortBlock;
		}else{
			((GeoPoint)pointList.get(nPoint)).leftCenterPoint = (Point2D)centerPoint;
			((GeoPoint)pointList.get(nPoint)).leftViewPort = viewPortBlock;
		}
	}

	/**
	 * Obtiene el extent de la mini imagen
	 * @param nPoint Punto de la lista
	 * @return
	 */
	public ViewPort getMiniExtent(int nPoint, boolean isRight){
		if(isRight)
			return ((GeoPoint)pointList.get(nPoint)).rightViewPort;
		else
			return ((GeoPoint)pointList.get(nPoint)).leftViewPort;
	}

	/**
	 * Obtiene el punto central de la mini imagen
	 * @param nPoint Punto de la lista
	 * @return
	 */
	public Point2D getCenterPoint(int nPoint, boolean isRight){
		if(isRight)
			return ((GeoPoint)pointList.get(nPoint)).rightCenterPoint;
		else
			return ((GeoPoint)pointList.get(nPoint)).leftCenterPoint;
	}

	/**
	 * Obtiene le ltimo punto de la lista
	 * @return El ltimo punto georreferenciado de la lista
	 */
	public GeoPoint getLastPoint(){
		return (GeoPoint)pointList.get(pointList.size());
	}

	/**
	 * Obtiene el array de puntos
	 * @return
	 */
	public ArrayList getListPoint(){
		return pointList;
	}

	/**
     * Obtiene el gestor de puntos
     * @return PointManager
     */
    public PointManager getPointManager() {
		return pointManager;
	}

    /**
     * Asigna el gestor de puntos
     * @param PointManager
     */
    public void setPointManager(PointManager pointManager) {
		this.pointManager = pointManager;
	}

    /**
     * Obtiene la capa de la imagen a georreferenciar
     * @return FLyrGeoRaster
     */
    public FLyrGeoRaster getLyrGeoRaster() {
		return lyrGeoRaster;
	}

    /**
	 * Asigna la capa de la imagen a georreferenciar
	 * @param flyGeoRaster
	 */
	public void setLyrGeoRaster(FLyrGeoRaster lyGeoRaster){
		this.lyrGeoRaster = lyGeoRaster;
		if(geoPointPersistence == null)
			geoPointPersistence = new GeoPointPersistence(this);
	}

	/**
	 * Asigna el flag de visibilidad a true o false. Si est a true la capa
	 * es visible normalmente y si est a false no se mostrar
	 * @param visible flag de visibilidad
	 */
	public void setVisible(boolean visible){
		this.lyrVisible = visible;
	}

	/**
	 * Obtiene el flag de visibilidad true o false. Si est a true la capa
	 * es visible normalmente y si est a false no se muestra
	 * @return flag de visibilidad
	 */
	public boolean isVisible(){
		return this.lyrVisible;
	}

	/**
	 * Asigna el flag de visibilidad a true o false. Si est a true la capa
	 * es visible normalmente y si est a false no se mostrar. Este flag solo
	 * afecta a la vista de gvSIG. Las mini-imagenes seguiran mostrandose.
	 * @param visible flag de visibilidad
	 */
	public void setHideViewOnly(boolean hideViewOnly){
		lyrHideViewOnly = hideViewOnly;
	}

	/**
	 * Obtiene el flag de visibilidad true o false. Si est a true la capa
	 * es visible normalmente y si est a false no se muestra. Este flag solo
	 * afecta a la vista de gvSIG. Las mini-imagenes seguiran mostrandose.
	 * @return flag de visibilidad
	 */
	public boolean isHideViewOnly(){
		return lyrHideViewOnly;
	}

	/**
	 * Obtiene la lista de puntos transformados. La lista de puntos transformada es la lista de puntos
	 * a la que se le ha aplicado la transformacin afn en base a las muestras tomadas.
	 * Esta lista de puntos transformados hay que mantenerla actualizada cada vez que se hace un
	 * update de un punto ya se usa para calcular los errores.
	 * @return Lista de puntos transformada.
	 */
	public Point2D[] getTransformPointList() {
		return transformPointList;
	}

	/**
	 * Asigna la lista de puntos transformada. La lista de puntos transformada es la lista de puntos
	 * a la que se le ha aplicado la transformacin afn en base a las muestras tomadas.
	 * Esta lista de puntos transformados hay que mantenerla actualizada cada vez que se hace un
	 * update de un punto ya se usa para calcular los errores.
	 * @param transformPointList Lista de puntos transformada.
	 */
	public void setTransformPointList(Point2D[] transformPointList) {
		this.transformPointList = transformPointList;
	}

	/**
	 * Asigna el objeto que gestiona la persistencia
	 */
	public void setPersistence(IPersistence geoPointPersistence) {
		if(this.geoPointPersistence == null)
			this.geoPointPersistence = geoPointPersistence;
	}

	/**
	 * Obtiene el valor del flag que dice si se dibuja la cruz azul
	 * @return
	 */
	public boolean isWithoutBlue() {
		return withoutBlue;
	}

	/**
	 * Asigna valor al flag que dice si se dibuja la cruz azul. Ser true si no queremos
	 * dibujarla y false si queremos.
	 * @param withoutBlue
	 */
	public void setWithoutBlue(boolean withoutBlue) {
		this.withoutBlue = withoutBlue;
	}

	/**
	 * Obtiene el valor del flag que dice si se dibuja la cruz roja
	 * @return
	 */
	public boolean isWithoutRed() {
		return withoutRed;
	}

	/**
	 * Asigna valor al flag que dice si se dibuja la cruz roja. Ser true si no queremos
	 * dibujarla y false si queremos.
	 * @param withoutRed
	 */
	public void setWithoutRed(boolean withoutRed) {
		this.withoutRed = withoutRed;
	}
	//**********************End Setters & Getters*********************
}
