/* gvSIG. Sistema de Informacin Geogrfica de la Generalitat Valenciana
 *
 * Copyright (C) 2004 IVER T.I. and Generalitat Valenciana.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,USA.
 *
 * For more information, contact:
 *
 *  Generalitat Valenciana
 *   Conselleria d'Infraestructures i Transport
 *   Av. Blasco Ibez, 50
 *   46010 VALENCIA
 *   SPAIN
 *
 *      +34 963862235
 *   gvsig@gva.es
 *      www.gvsig.gva.es
 *
 *    or
 *
 *   IVER T.I. S.A
 *   Salamanca 50
 *   46005 Valencia
 *   Spain
 *
 *   +34 963163400
 *   dac@iver.es
 */

package org.gvsig.gui.beans.comboBoxItemsSeeker;

import java.util.Vector;

import javax.swing.DefaultComboBoxModel;

import org.apache.log4j.Logger;

/**
 * Abstract class with some abstract methods that have to implement the classes which inherit this; and with other implementented methods for
 *    initialize the ComboBoxItemsSeekerConfigurableModel and do a binary seach if its needed.
 * This class represents a Model of the JComboBoxItemsSeekerConfigurable, extends DefaultComboBoxModel, and can be configurated the items that returns.
 * In the MVC (Model-View-Control) pattern, this class represents the Model -> this means that this class returns the data (items), in this case
 *     according its configuration and the selected item.
 * It could be necessary obtein information about the model of the parent of this class: DefaultComboBoxModel -> there are some methods that allow this.
 *     
 * The model that represents this class, can be configurable with 3 flags/attributes:
 *    - Start Behavior -> Configure how return items at beginning (always returns all items). 
 *       + MAINTAIN_ORIGINAL_POSITION_START -> Returns items in the same order as DefaultComboBoxModel (according they were introduced).
 *       + ORDERED_START -> Returns items ordered.
 *       + DISORDERED_START -> Returns items disordered.
 *    - Search Behavior -> Configure how return items when it's searching an item.
 *       + MAINTAIN_ORIGINAL_POSITION_ALL_ITEMS_SEARCH -> Returns all items in the same order as DefaultComboBoxModel (according they were introduced).
 *       + ORDERED_ALL_ITEMS_SEARCH -> Returns all items ordered
 *       + DISORDERED_ALL_ITEMS_SEARCH -> Returns all items disordered
 *       + MAINTAIN_ORIGINAL_POSITION_DYNAMIC_SEARCH -> Retuns items that their String value starts with a chain of characters introduced. Items are returned in
 *           the same order as DefaultComboBoxModel (according they were introduced).
 *       + ORDERED_DYNAMIC_SEARCH -> Retuns items that their String value starts with a chain of characters introduced. Items are returned ordered.
 *       + DISORDERED_DYNAMIC_SEARCH -> Retuns items that their String value starts with a chain of characters introduced. Items are returned disordered.
 *   - Case Sensitive -> Discriminates capital letters from small letters or not when seeks items
 *   	 + true -> discriminates capital letters from small letters when seeks items
 *       + false -> don't discriminates capital letters from small letters when seeks items
 *       
 * By default the configuration is:
 *     - Start_Behavior = MAINTAIN_ORIGINAL_POSITION_START
 *     - Search_Behavior = ORDERED_DYNAMIC_SEARCH
 *     - Case_Sensitive = false
 * 
 * @author Pablo Piqueras Bartolom (p_queras@hotmail.com)
 */
public abstract class AbstractDefaultComboBoxItemsSeekerConfigurableModel extends DefaultComboBoxModel {
	// INNER MODEL STRUCTURES
	protected Object itemsSearchesShow[];
	// END INNER MODEL STRUCTURES
	
	// TRACE - DEBUG
	protected Logger logger = Logger.getLogger(AbstractDefaultComboBoxItemsSeekerConfigurableModel.class.getClass());
	// END TRACE - DEBUG
	
	// CONTANTS FOR CONFIGURE THE BEHAVIOR
	public static final int MAINTAIN_ORIGINAL_POSITION_START = 0;
	public static final int ORDERED_START = 1;
	public static final int DISORDERED_START = 2;

	public static final int MAINTAIN_ORIGINAL_POSITION_ALL_ITEMS_SEARCH = 3;
	public static final int ORDERED_ALL_ITEMS_SEARCH = 4;
	public static final int DISORDERED_ALL_ITEMS_SEARCH = 5;
	public static final int MAINTAIN_ORIGINAL_POSITION_DYNAMIC_SEARCH = 6;
	public static final int ORDERED_DYNAMIC_SEARCH = 7;
	public static final int DISORDERED_DYNAMIC_SEARCH = 8;
	
	public static final int DEFAULT_START_BEHAVIOR_CONFIGURATION = MAINTAIN_ORIGINAL_POSITION_START;
	public static final int DEFAULT_SEARCH_BEHAVIOR_CONFIGURATION = ORDERED_DYNAMIC_SEARCH;
	public static final boolean DEFAULT_CASE_SENSITIVE_CONFIGURATION = false;
	// END CONTANTS FOR CONFIGURE THE BEHAVIOR
	
	// CONFIGURATION FLAGS
	protected int startBehavior;
	protected int searchBehavior;
	protected boolean caseSensitive_Flag;
	// END FLAGS

	// FLAGS AND ATTRIBUTES FOR THE CURRENT STATE
	protected int stateAllItems;	
	protected final int STATE_DEFAULT = 10;
	protected final int STATE_ITEMS_ARE_ORDERED = 11;
	protected final int STATE_ITEMS_ARE_DISORDERED = 12;
	// FLAGS AND ATTRIBUTES FOR THE CURRENT STATE	
	
	// OTHER FLAGS
	protected boolean startState;
	protected boolean dynamicSearchDone;
	protected boolean itemsInTheArray;
	// END OTHER FLAGS

	// ATTRIBUTES FOR DYNAMIC CASES
	protected String writtenText;
	// END ATTRIBUTES FOR DYNAMIC CASES
	
	// OTHER OBJECTS
	protected Object moreSimilarItemToLastWrittenText;
	// OTHER OBJECTS
	
	/** 
	 * @see DefaultComboBoxModel#DefaultComboBoxModel()
	 */
	public AbstractDefaultComboBoxItemsSeekerConfigurableModel() {
		super();
	}
	
	/** 
	 * @see DefaultComboBoxModel#DefaultComboBoxModel(Object[])
	 */
	public AbstractDefaultComboBoxItemsSeekerConfigurableModel(Object[] items) {		
		super(items);
	}

	/**
	 * @see DefaultComboBoxModel#DefaultComboBoxModel(Vector)
	 */
	public AbstractDefaultComboBoxItemsSeekerConfigurableModel(Vector objects) {
		super(objects);
	}
	
	/**
	 * This method sets the start values of inner attributes and creates the necessary inner objects
	 */
	protected void initialize() {
		this.writtenText = "";
		this.setDefaultBehaviorFlagsConfiguration();
		this.startState = true;
		this.stateAllItems = STATE_DEFAULT;

		this.dynamicSearchDone = false; 
		this.itemsInTheArray = false;
	}
	
	/**
	 * Method that sets the default values of the behavior flags
	 */
	protected void setDefaultBehaviorFlagsConfiguration() {
		// Set the default values of the flags
		this.startBehavior = DEFAULT_START_BEHAVIOR_CONFIGURATION;
		this.searchBehavior = DEFAULT_SEARCH_BEHAVIOR_CONFIGURATION;
		this.caseSensitive_Flag = DEFAULT_CASE_SENSITIVE_CONFIGURATION;
	}
	
	////// METHODS FOR THE BEHAVIOR FLAGS  //////
	
	/**
	 * This method tests the configuration of the behavior flags and returns true or false if its ok or not
	 * 
	 * @return boolean True if the configuration of the flags is oks, false if not
	 */
	public boolean testFlagsConfigurationOK() {
		if ((this.startBehavior < 0) || (this.startBehavior > 2))
			return false;
		
		if ((this.searchBehavior < 3) || (this.searchBehavior > 8))
			return false;
		
		return true;
	}

	/**
	 * Retuns the 'start_Behavior' configuration of this model. Configuration values are:
     *    + MAINTAIN_ORIGINAL_POSITION_START -> Returns items in the same order as DefaultComboBoxModel (according they were introduced).
     *    + ORDERED_START -> Returns items ordered.
     *    + DISORDERED_START -> Returns items disordered.
	 * 
	 * @return 'start_Behavior' configuration
	 */
	public int getStartBehavior() {
		return this.startBehavior;
	}
	
	/**
	 * Sets the 'start_Behavior' configuration for this model. Configuration values are:
     *    + MAINTAIN_ORIGINAL_POSITION_START -> Returns items in the same order as DefaultComboBoxModel (according they were introduced).
     *    + ORDERED_START -> Returns items ordered.
     *    + DISORDERED_START -> Returns items disordered.
	 * 
	 * @oaram An integer value for 'start_Behavior' configuration flag
	 */
	public void setStartBehavior(int start_Behavior) {
		this.startBehavior = start_Behavior;		
	}

	/**
	 * Retuns the 'search_Behavior' configuration of this model. Configuration values are:
     *    + MAINTAIN_ORIGINAL_POSITION_ALL_ITEMS_SEARCH -> Returns all items in the same order as DefaultComboBoxModel (according they were introduced).
     *    + ORDERED_ALL_ITEMS_SEARCH -> Returns all items ordered
     *    + DISORDERED_ALL_ITEMS_SEARCH -> Returns all items disordered
     *    + MAINTAIN_ORIGINAL_POSITION_DYNAMIC_SEARCH -> Retuns items that their String value starts with a chain of characters introduced. Items are returned in
     *        the same order as DefaultComboBoxModel (according they were introduced).
     *    + ORDERED_DYNAMIC_SEARCH -> Retuns items that their String value starts with a chain of characters introduced. Items are returned ordered.
     *    + DISORDERED_DYNAMIC_SEARCH -> Retuns items that their String value starts with a chain of characters introduced. Items are returned disordered.
	 * 
	 * @return 'search_Behavior' configuration
	 */
	public int getSearchBehavior() {
		return this.searchBehavior;
	}

	/**
	 * Sets the 'search_Behavior' configuration for this model. Configuration values are:
     *   + MAINTAIN_ORIGINAL_POSITION_ALL_ITEMS_SEARCH -> Returns all items in the same order as DefaultComboBoxModel (according they were introduced).
     *   + ORDERED_ALL_ITEMS_SEARCH -> Returns all items ordered
     *   + DISORDERED_ALL_ITEMS_SEARCH -> Returns all items disordered
     *   + MAINTAIN_ORIGINAL_POSITION_DYNAMIC_SEARCH -> Retuns items that their String value starts with a chain of characters introduced. Items are returned in
     *       the same order as DefaultComboBoxModel (according they were introduced).
     *   + ORDERED_DYNAMIC_SEARCH -> Retuns items that their String value starts with a chain of characters introduced. Items are returned ordered.
     *   + DISORDERED_DYNAMIC_SEARCH -> Retuns items that their String value starts with a chain of characters introduced. Items are returned disordered.
	 * 
	 * @oaram An integer value for 'search_Behavior' configuration flag
	 */
	public void setSearchBehavior(int search_Behavior) {
		this.searchBehavior = search_Behavior;
	}

	/**
	 * Returns the 'case_Sensitive' configuration value of this model. Configuration values are: 
	 *   + true -> discriminates capital letters from small letters when seeks items
	 *   + false -> don't discriminates capital letters from small letters when seeks items
	 * 
	 * @return 'case_Sensitive' configuration
	 */
	public boolean isCaseSensitive_Flag() {
		return caseSensitive_Flag;
	}
	
	/**
	 * Sets the 'case_Sensitive' configuration value for this model. Configuration values are: 
	 *   + true -> discriminates capital letters from small letters when seeks items
	 *   + false -> don't discriminates capital letters from small letters when seeks items
	 * 
	 * @oaram A boolean value for 'case_Sensitive' configuration flag
	 */
	public void setCaseSensitive_Flag(boolean case_Sensitive) {
		this.caseSensitive_Flag = case_Sensitive;
	}

	////// END METHODS FOR THE BEHAVIOR FLAGS //////
	

	
	////// REIMPLEMENTATION OF SOME METHODS OF "DefaultComboBoxModel" TO ADAPT THE BEHAVIOR OF THIS MODEL //////	
	
	/*
	 *  (non-Javadoc)
	 * @see javax.swing.DefaultComboBoxModel#getSize()
	 */
	public abstract int getSize();	

	/*
	 *  (non-Javadoc)
	 * @see javax.swing.ListModel#getElementAt(int)
	 */
	public abstract Object getElementAt(int index);

	/*
	 *  (non-Javadoc)
	 * @see javax.swing.DefaultComboBoxModel#getIndexOf(java.lang.Object)
	 */
	public abstract int getIndexOf(Object anObject);
	
    /*
     *  (non-Javadoc)
     * @see javax.swing.ComboBoxModel#getSelectedItem()
     */
	public Object getSelectedItem() {
		return super.getSelectedItem();
	}
	
	/*
	 *  (non-Javadoc)
	 * @see javax.swing.MutableComboBoxModel#addElement(java.lang.Object)
	 */
	public abstract void addElement(Object anObject);
	
	/*
	 *  (non-Javadoc)
	 * @see javax.swing.MutableComboBoxModel#insertElementAt(java.lang.Object, int)
	 */
	public abstract void insertElementAt(Object anObject, int index);

	/*
	 *  (non-Javadoc)
	 * @see javax.swing.DefaultComboBoxModel#removeAllElements()
	 */
	public abstract void removeAllElements();
    
	/*
	 *  (non-Javadoc)
	 * @see javax.swing.MutableComboBoxModel#removeElement(java.lang.Object)
	 */
	public abstract void removeElement(Object anObject);
	
	/*
	 *  (non-Javadoc)
	 * @see javax.swing.MutableComboBoxModel#removeElementAt(int)
	 */
	public abstract void removeElementAt(int index);
	
	////// END REIMPLEMENTATION OF SOME METHODS OF "DefaultComboBoxModel" TO ADAPT THE BEHAVIOR OF THIS MODEL //////	

	
	
	////// USE OF SOME METHODS OF "DefaultComboBoxModel" //////	
	
	/**
     * @see DefaultComboBoxModel#getSize()
	 */
	public int getParentSize() {
		return super.getSize();
	}
	
	/**
	 * @see javax.swing.MutableComboBoxModel#addElement(java.lang.Object)
	 */
	protected void addElementToParent(Object anObject) {
		super.addElement(anObject);
	}
	
	/**
	 * @see javax.swing.ListModel#getElementAt(int)
	 */
	protected Object getParentElementAt(int index) {
		return super.getElementAt(index);
	}
	
	/**
	 * @see javax.swing.DefaultComboBoxModel#getIndexOf(java.lang.Object)
	 */
	protected int getParentIndexOf(Object anObject) {
		return super.getIndexOf(anObject);
	}
	
	/**
	 * @see javax.swing.MutableComboBoxModel#insertElementAt(java.lang.Object, int)
	 */
	protected void insertToParentElementAt(Object anObject, int index) {
		super.insertElementAt(anObject, index);
	}
	
	/**
	 * @see javax.swing.DefaultComboBoxModel#removeAllElements()
	 */
	protected void removeAllElementsOfParent() {
		super.removeAllElements();
	}
	
	/**
	 * @see javax.swing.MutableComboBoxModel#removeElement(java.lang.Object)
	 */
	protected void removeParentElement(Object anObject) {
		super.removeElement(anObject);
	}
	
	/**
	 * @see javax.swing.MutableComboBoxModel#removeElementAt(int)
	 */
	protected void removeParentElementAt(int index) {
		// TODO Auto-generated method stub
		super.removeElementAt(index);
	}

	////// END USE OF SOME METHODS OF "DefaultComboBoxModel" //////	

	
	
	////// OTHER METHODS //////
	
	/**
	 * Returns all items that are stored in the model.
	 * Caution! : this operation could be slow
	 */
	public abstract Vector getParentAllData();
	
	/**
	 * Returns the first item added
	 *  
	 * @return An item
	 */
	public Object getFirstItemAdded() {
	
		return super.getElementAt(0);
	}
	
	/**
	 * Sets a text for seek items
	 */
	public void setWrittenText(String text) {
		this.writtenText = text;
		this.reset();
	}
	
	/**
	 * Returns the item of the model that its text value is more similar to the written text, or null
	 * 
	 * @return An Item or null
	 */
	public Object getMoreSimilarItemToLastWrittenText() {
		int size = getSize();
		
		if (size == 0)
			return null;
		else {
			moreSimilarItemToLastWrittenText = getElementAt(0);
			return moreSimilarItemToLastWrittenText;
		}
	}
	
	/**
	 * Indicates to this model that now has to be or not in the 'start' state.
	 * If 'flagValue' is true:  items returned will be according the 
	 *    start-behavior configuration.
	 * 
	 * @param flagValue A boolean value.
	 */
	public void setStartStateFlag(boolean flagValue) {
		this.startState = flagValue;
		this.reset();
	}
	
	/**
	 * Returns if now the model is in the 'start' state or not
	 * 
	 * @return True or false if now the model is in the 'start' state or not
	 */
	public boolean isStartStateFlag() {
		return this.startState;
	}
	
	/**
	 * Resets some inner configuration flags to their default value
	 */
	protected void reset() {
		if (this.stateAllItems != STATE_DEFAULT)
		{
			this.stateAllItems = STATE_DEFAULT;
			this.dynamicSearchDone = false;
			this.itemsInTheArray = false;
		}
	}
	
	/**
	 * Returns true of false if the configuration of the search is dynamic search or not
	 * @return True of false if the configuration of the search is dynamic search or not
	 */
	public boolean isDynamicSearch() {
		if ((this.searchBehavior > 5) && (this.searchBehavior < 9))
			return true;
		else
			return false;
	}
	
	/**
	 * With this method the model does a dynamic items seek
	 */
	protected abstract void seekDynamicItems();	

	/**
	 * Returns if the items returned are disordered
	 * 
	 * @return If the items returned are disordered
	 */
	public boolean nowItemsAreDisordered() {
		if (this.startState)
		{
			if (getStartBehavior() == DISORDERED_START)
				return true;
		}
		else
		{
			if ((getSearchBehavior() == DISORDERED_DYNAMIC_SEARCH) || (getSearchBehavior() == DISORDERED_ALL_ITEMS_SEARCH))
				return true;
		}
		
		return false;
	}
	
	/**
	 * Gets an array with all items that will return/show the model. It's used in dynamic searches.
	 * 
	 * @param search An special binary of items search object.
	 * @param lowIndex Low index of the rank of items to extract from the search object.
	 * @param hightIndex Hight index of the rank of items to extract from the search object.
	 */
	protected abstract void getDynamicItemsSearch(BinaryRankOfItemsSeek search, int lowIndex, int hightIndex);	
	
	/**
	 * Copies items from a search into an array
	 */
	protected abstract void setItemsCopyToArray();
	
	/**
	 * This method disorders the items and stores them in an array
	 */
	protected abstract void disorderItemsCopy();
	
	////// END OTHER METHODS //////
	
	
	/**
	 * Inner class for items seek.
	 * Considers that lowcases and uppercases words/senteces could be in differents ranks.
	 * Has some modifications of bynary-search-algorithm for obtein the low and high indexes of the lowcase and uppercase ranks. Not always there
	 *    are the two ranks, it depends on the 'case-sensitive' flag.
	 * 
     * @author Pablo Piqueras Bartolom (p_queras@hotmail.com)
	 */
	protected class BinaryRankOfItemsSeek {
		private int lowIndex;
		private int hightIndex;
		private int lowIndexSubSet2; // Only used for not caseSensitive case
		private int hightIndexSubSet2; // Only used for not caseSensitive case
		
		/**
		 * Default constructor
		 */
		public BinaryRankOfItemsSeek() {
			// By default -> no search has been made
			this.lowIndex = -1;
			this.hightIndex = -1;
			this.lowIndexSubSet2 = -1;
			this.hightIndexSubSet2 = -1;
		}
	
		/**
		 * This method has the algoritms for obtein the low and high rank indexes for 2 ranks: one for items that start in a capital letter
		 *    and other which start in a small letter. It also could use different algorithm/s if it's needed differenciate uppercases and
		 *    lowercases or not. If its needed to differenciate: there are 2 ranks, else there is only one rank.
		 * Uses the beginning of the word/sentece ('writtenText' attribute) for match its start with the start of all items ordered in a/some
		 *     binary algorithm, and obtein the indexes of the 2 ranks.
		 */
		public void binaryRankSeek() {
			boolean continueInLoop = true;
			int mediumIndex;
			int maxNumberOfIterations = (int) MathExtensionReduced.log2(itemsSearchesShow.length);
			logger.debug("Number of iterations: " + maxNumberOfIterations);
			int currentIteration = 0;
				
			this.lowIndex = 0;
			this.hightIndex = itemsSearchesShow.length - 1;
			
			// First -> Find the first item that match its start (This is an iterational version of the binarySearch algorithm)
			if (isCaseSensitive_Flag())
			{
				// Iterational version of a binary search of the first item that it start matches with a text
				while (continueInLoop)
				{
					// If can't continue searching (item hasn't been found) (With this condition we avoid infinites loops)
					if (currentIteration > maxNumberOfIterations)
					{
						this.lowIndex = this.hightIndex + 1; // Force not show any item after
						continueInLoop = false;							
					}
					else
					{
						mediumIndex = (int) Math.floor((this.lowIndex + this.hightIndex) / 2);
											
						// If has find the first item that mathes its start
						if (itemsSearchesShow[mediumIndex].toString().startsWith(writtenText))
						{
							this.lowIndex = this.hightIndex = mediumIndex;
							continueInLoop = false;
						}
						else
						{
							// Reduce the rank of search
							if (itemsSearchesShow[mediumIndex].toString().compareTo(writtenText) > 0)
								this.hightIndex = mediumIndex - 1;							
							else
								this.lowIndex = mediumIndex + 1;
							
							// If can't continue searching (item hasn't been found)
							if (this.lowIndex > this.hightIndex)
								continueInLoop = false;							
							else
								currentIteration ++;
						}
					}
				}
				
				logger.debug("BINARY-SEARCH: Low-Index: " + this.lowIndex + "; Hight-Index: " + this.hightIndex);
			}
			else
			{	
				// Elements are ordered in this order -> First UpperCases, Second Simbols, Third LowerCases (left -> right character priority)
				
				// Get The first SubSet
				// Iterational version of a binary search of the first item that its start matches with a text
				while (continueInLoop)
				{
					// If can't continue searching (item hasn't been found) (With this condition we avoid infinites loops)
					if (currentIteration > maxNumberOfIterations)
					{
						this.lowIndex = this.hightIndex + 1; // Force not show any item after
						continueInLoop = false;							
					}
					else
					{
						mediumIndex = (int) Math.floor((this.lowIndex + this.hightIndex) / 2);
											
						// If has find the first item that mathes its start
						if (itemsSearchesShow[mediumIndex].toString().toUpperCase().startsWith(writtenText.toUpperCase()))
						{
							this.lowIndex = this.hightIndex = mediumIndex;
							continueInLoop = false;
						}
						else
						{
							// Reduce the rank of search
							if (itemsSearchesShow[mediumIndex].toString().compareToIgnoreCase(writtenText) > 0) // da problemas
								this.hightIndex = mediumIndex - 1;							
							else
								this.lowIndex = mediumIndex + 1;
							
							// If can't continue searching (item hasn't been found)
							if (this.lowIndex > this.hightIndex)
								continueInLoop = false;							
							else
								currentIteration ++;
						}
					}
				}
			
				logger.debug("BINARY-SEARCH-FirstSubSet: Low-Index: " + this.lowIndex + "; Hight-Index: " + this.hightIndex);
				
				
				// THE SECOND SUBSET HAS TO SEEK OLY THE ITEMS THAT START IN CAPITAL LETTERS
				// Get The second SubSet
				continueInLoop = true;
				currentIteration = 0;
				this.lowIndexSubSet2 = 0;
				this.hightIndexSubSet2 = itemsSearchesShow.length - 1;

				// Iterational version of a binary search of the first item that its start matches with a text
				// In this case tries to find a subset of items that ther string value start with uppercase
				while (continueInLoop)
				{
					// If can't continue searching (item hasn't been found) (With this condition we avoid infinites loops)
					if (currentIteration > maxNumberOfIterations)
					{
						this.lowIndexSubSet2 = this.hightIndexSubSet2 + 1; // Force not show any item after
						continueInLoop = false;							
					}
					else
					{
						mediumIndex = (int) Math.floor((this.lowIndexSubSet2 + this.hightIndexSubSet2) / 2);
											
						// If has find the first item that mathes its start
						if (itemsSearchesShow[mediumIndex].toString().toUpperCase().startsWith(writtenText.toUpperCase()))
						{
							this.lowIndexSubSet2 = this.hightIndexSubSet2 = mediumIndex;
							continueInLoop = false;
						}
						else
						{
							// Reduce the rank of search
							if (itemsSearchesShow[mediumIndex].toString().compareTo(writtenText.toUpperCase()) > 0) // da problemas
								this.hightIndexSubSet2 = mediumIndex - 1;							
							else
								this.lowIndexSubSet2 = mediumIndex + 1;
							
							// If can't continue searching (item hasn't been found)
							if (this.lowIndexSubSet2 > this.hightIndexSubSet2)
								continueInLoop = false;							
							else
								currentIteration ++;
						}
					}
				}
				
				logger.debug("BINARY-SEARCH-SecondSubSet: Low-Index: " + this.lowIndexSubSet2 + "; Hight-Index: " + this.hightIndexSubSet2);
				
				// Force not do another search of rank of items
				if ((lowIndex == lowIndexSubSet2) && (hightIndex == hightIndexSubSet2))
					lowIndexSubSet2 = hightIndexSubSet2 + 1;
			}
					
			
			// Second -> if indexes are the same -> search the rank of items (items are ordered in the 'itemsSearchesShow' array
			if (this.lowIndex == this.hightIndex)
			{				
				
				if (isCaseSensitive_Flag())
				{
					// Seek down
					this.lowIndex --;
					if (this.lowIndex > -1)
						continueInLoop = true;
					else
						continueInLoop = false;

					while (continueInLoop)
					{
						if (!itemsSearchesShow[this.lowIndex].toString().startsWith(writtenText))
							continueInLoop = false; // Because items are sorted in the array 'itemsSearchesShow'
						else
						{
							this.lowIndex --;
							if (this.lowIndex < 0) // Because the item at the 0 position has already been verified
								continueInLoop = false;
						}
					}
					this.lowIndex++;
					
					// Seek up
					this.hightIndex ++;
					if (this.hightIndex < itemsSearchesShow.length)
						continueInLoop = true;
					else
						continueInLoop = false;
					
					while (continueInLoop)
					{
						if (!itemsSearchesShow[this.hightIndex].toString().startsWith(writtenText)) 
							continueInLoop = false; // Because items are sorted in the array 'itemsSearchesShow'
						else
						{
							this.hightIndex ++;
							if (this.hightIndex >= itemsSearchesShow.length) // Because the item at the 0 position has already been verified
								continueInLoop = false;
						}
					}
					this.hightIndex --;				
				}
				else
				{
					// Seek down
					this.lowIndex --;
					if (this.lowIndex > -1)
						continueInLoop = true;
					else
						continueInLoop = false;

					while (continueInLoop)
					{
						if (!itemsSearchesShow[this.lowIndex].toString().toUpperCase().startsWith(writtenText.toUpperCase()))
							continueInLoop = false; // Because items are sorted in the array 'itemsSearchesShow'
						else
						{
							this.lowIndex --;
							if (this.lowIndex < 0) // Because the item at the 0 position has already been verified
								continueInLoop = false;
						}
					}
					this.lowIndex++;
					
					// Seek up
					this.hightIndex ++;
					if (this.hightIndex < itemsSearchesShow.length)
						continueInLoop = true;
					else
						continueInLoop = false;
					
					while (continueInLoop)
					{
						if (!itemsSearchesShow[this.hightIndex].toString().toUpperCase().startsWith(writtenText.toUpperCase()))
							continueInLoop = false; // Because items are sorted in the array 'itemsSearchesShow'
						else
						{
							this.hightIndex ++;
							if (this.hightIndex >= itemsSearchesShow.length) // Because the item at the 0 position has already been verified
								continueInLoop = false;
						}
					}
					this.hightIndex --;					
				}				
			}
			logger.debug("Low-Index: " + this.lowIndex + "; Hight-Index: " + this.hightIndex);
			
			

			// This executed only for cases which isn't enabled case sensitive
			if (!isCaseSensitive_Flag())
			{
				// Second -> if indexes are the same -> search the rank of items (items are ordered in the 'itemsSearchesShow' array
				if (this.lowIndexSubSet2 == this.hightIndexSubSet2)
				{				

					// Seek down
					this.lowIndexSubSet2 --;
					if (this.lowIndexSubSet2 > -1)
						continueInLoop = true;
					else
						continueInLoop = false;

					while (continueInLoop)
					{
						if (!itemsSearchesShow[this.lowIndexSubSet2].toString().toUpperCase().startsWith(writtenText.toUpperCase()))
							continueInLoop = false; // Because items are sorted in the array 'itemsSearchesShow'
						else
						{
							this.lowIndexSubSet2 --;
							if (this.lowIndexSubSet2 < 0) // Because the item at the 0 position has already been verified
								continueInLoop = false;
						}
					}
					this.lowIndexSubSet2++;
					
					// Seek up
					this.hightIndexSubSet2 ++;
					if (this.hightIndexSubSet2 < itemsSearchesShow.length)
						continueInLoop = true;
					else
						continueInLoop = false;
					
					while (continueInLoop)
					{
						if (!itemsSearchesShow[this.hightIndexSubSet2].toString().toUpperCase().startsWith(writtenText.toUpperCase()))
							continueInLoop = false; // Because items are sorted in the array 'itemsSearchesShow'
						else
						{
							this.hightIndexSubSet2 ++;
							if (this.hightIndexSubSet2 >= itemsSearchesShow.length) // Because the item at the 0 position has already been verified
								continueInLoop = false;
						}
					}
					this.hightIndexSubSet2 --;
				}				
				
				logger.debug("Low-Index: " + this.lowIndexSubSet2 + "; Hight-Index: " + this.hightIndexSubSet2);
			}
		}
	
		/**
		 * Returns the low index of the first rank.
		 * 
		 * @return Low index value.
		 */
		public int getLowIndex() {
			return this.lowIndex;
		}
		
		/**
		 * Returns the hight index of the first rank.
		 * 
		 * @return Hight index value
		 */
		public int getHightIndex() {
			return this.hightIndex;
		}

		/**
		 * Returns the low index of the second rank. (It's used only when the flag 'case-sensitive' is true.) 
		 * 
		 * @return Low index value.
		 */
		public int getLowIndexSecondSubSet() {
			return this.lowIndexSubSet2;
		}
		
		/**
		* Returns the hight index of the second rank. (It's used only when the flag 'case-sensitive' is true.) 
		* 
		* @return Hight index value.
		*/
		public int getHightIndexSecondSubSet() {
			return this.hightIndexSubSet2;
		}
	}
}
