package org.agil.core.spatialindex;

import com.vividsolutions.jts.geom.*;
import java.io.*;
import java.util.*;

/**
 *  To create a <i>new</i> RTree use the first two constructors. You must
 *  specify the dimension, the fill factor as a float between 0 and 0.5 (0 to
 *  50% capacity) and the variant of the RTree which is one of:
 *  <ul>
 *    <li> RTREE_QUADRATIC</li>
 *    <li> RTREE_LINEAR</li>
 *  </ul>
 *  The first constructor creates by default a new memory resident page file.
 *  The second constructor takes the page file as an argument. If the given
 *  page file is not empty, then all data are deleted. <p>
 *
 *  The third constructor initializes the RTree from the given page file.
 *  Thus, you may store the RTree into a persistent page file and reload it at
 *  any time. <p>
 *
 *  Created: Tue May 18 12:57:35 1999 <p>
 *
 *
 *
 *@author     Marios Hadjieleftheriou
 *@version    2
 */
public class RTree {
	/**
	 *  Page file where data is stored.
	 */
	protected PageFile file = null;
	/**
	 *  RTree variant used.
	 */
	protected int treeType = -1;
	/**
	 *  Minimum number of entries per node.
	 */
	protected int minimumLoad = -1;
	/**
	 *  Object for storing several statistics.
	 */
	protected Statistics stats = new Statistics();
	//    public final String version = "2";
	//    public final String date = "December 31, 2000";

	/**
	 *  Nombre del fichero que almacena la estructura y datos del RTree
	 */
	String treeName;
	/**
	 *  Para iterator
	 */
	private boolean hasNext;
	private Enumeration nodos;
	// nodos devueltos por la busqueda
	private AbstractNode nodo;
	// nodo por el que va el iterator
	private int num_nodo;
	// n del nodo por el que va el iterator
	private int num_elto;
	// elto. del nodo por el que va el iterator
	private HyperCube mer;
	private boolean isOpen = false;
	/**
	 *  static identifier used for the parent of the root node.
	 */
	public static final int NIL = -1;
	/**
	 *  Available RTree variants.
	 */
	public static final int RTREE_LINEAR = 0;
	/**
	 *  Description of the Field
	 */
	public static final int RTREE_QUADRATIC = 1;
	/**
	 *  Description of the Field
	 */
	public static final int RSTAR = 2;
	// mer buscado
			/**
	 *  Description of the Field
	 */
	public static final int NO_MORE_ELEMENTS = -1;


	/**
	 *  Constructor for the RTree object
	 *
	 *@param  directory  Description of Parameter
	 *@param  name       Description of Parameter
	 */
	public RTree(String directory, String name) {
		treeName = directory + File.separator + "rtree_" + name + ".idx";
		hasNext = false;
	}


	/**
	 *  Devuelve el nombre del fichero que contiene los datos y estructura del
	 *  arbol.
	 *
	 *@return    El nombre del fichero
	 */
	public String getFileName() {
		return treeName;
	}


	/**
	 *  Retruns the maximun capacity of each Node.
	 *
	 *@return    The NodeCapacity value
	 */
	public int getNodeCapacity() {
		return file.nodeCapacity;
	}


	/**
	 *  Returns the percentage between 0 and 0.5, used to calculate minimum
	 *  number of entries present in each node.
	 *
	 *@return    The FillFactor value
	 */
	public float getFillFactor() {
		return file.fillFactor;
	}


	/**
	 *  Returns the minimum number of entries that a node must contain at any
	 *  time.
	 *
	 *@return    The MinimumLoad value
	 */
	public int getMinimumLoad() {
		return minimumLoad;
	}


	/**
	 *  Returns the data dimension.
	 *
	 *@return    The Dimension value
	 */
	public int getDimension() {
		return file.dimension;
	}


	/**
	 *  Returns the page length.
	 *
	 *@return    The PageSize value
	 */
	public int getPageSize() {
		return file.pageSize;
	}


	/**
	 *  Returns the level of the root Node, which signifies the level of the
	 *  whole tree. Loads one page into main memory.
	 *
	 *@return    The TreeLevel value
	 */
	public int getTreeLevel() {
		return file.readNode(0).getLevel();
	}


	/**
	 *  Returns the RTree variant used.
	 *
	 *@return    The TreeType value
	 */
	public int getTreeType() {
		return treeType;
	}



	/**
	 *  Devuelve el MER global de todos los nodos de este RTree.
	 *
	 *@return    MER global del RTree o null si no hay nodos.
	 */
	public Envelope getMER() {
		AbstractNode root = file.readNode(0);
		HyperCube hypercube = root.getNodeMbb();
		if (hypercube == null) {
			return null;
		}
		else {
			Point p1 = hypercube.getLowPoint();
			Point p2 = hypercube.getHighPoint();
//			return new Envelope(p1.getFloatCoordinate(0), p1.getFloatCoordinate(1), p2.getFloatCoordinate(0), p2.getFloatCoordinate(1));
			return new Envelope(p1.getFloatCoordinate(0), p2.getFloatCoordinate(0), p1.getFloatCoordinate(1), p2.getFloatCoordinate(1));
		}
	}


	/**
	 *  Creates a memory resident RTree with an empty root node which is stored
	 *  in page 0 and has a parent identifier equal to NIL. A new memory page
	 *  file is created and used as a storage manager, by default.
	 *
	 */

	/*
	 *  public RTree(int dimension, float fillFactor, int capacity, int type) {
	 *  this(dimension, fillFactor, capacity, new MemoryPageFile(), type);
	 *  }
	 */
	/**
	 *  Creates a new RTree with an empty root node which is stored in page 0 and
	 *  has a parent identifier equal to NIL. The given page file is used for
	 *  storing the rtree nodes. If the page file is not empty, than all entries
	 *  will be overwritten.
	 *
	 *@param  dimension        The data dimension.
	 *@param  fillFactor       A percentage between 0 and 0.5. Used to calculate
	 *      minimum number of entries present in each node.
	 *@param  capacity         The maximun number of entries that each node can
	 *      hold.
	 *@param  cacheSize        Description of Parameter
	 *@param  type             The RTree variant to use.
	 *@exception  IOException  Description of Exception
	 */
	public synchronized void create(int dimension, float fillFactor, int capacity, int cacheSize, int type) throws IOException {
		if (isOpen) {
			close();
		}
		file = new CachedPersistentPageFile(treeName, cacheSize, false);
		if (dimension <= 1) {
			throw new IllegalArgumentException("Dimension must be larger than 1.");
		}
		if (fillFactor < 0 || fillFactor > 0.5) {
			throw new IllegalArgumentException("Fill factor must be between 0 and 0.5.");
		}
		if (capacity <= 1) {
			throw new IllegalArgumentException("Capacity must be larger than 1.");
		}
		if (type != RTREE_QUADRATIC && type != RTREE_LINEAR) {
			throw new IllegalArgumentException("Invalid tree type.");
		}
		if (file.tree != null) {
			throw new IllegalArgumentException("PageFile already in use by another rtree instance.");
		}
		this.treeType = type;
		this.file = file;
		file.initialize(this, dimension, fillFactor, capacity);
		minimumLoad = Math.round(getNodeCapacity() * getFillFactor());
		if (minimumLoad < 2) {
			minimumLoad = 2;
		}
		Leaf root = new Leaf(this, NIL, 0);
		file.writeNode(root);
		isOpen = true;
	}


	/**
	 *  Crea con valores por defecto
	 *
	 *@exception  IOException  Description of Exception
	 */
	public void create() throws IOException {
		create(2, (float) 0.40, 25, 1024, RTREE_QUADRATIC);
	}


	/**
	 *  Creates an rtree from an already initialized page file, probably stored
	 *  into persistent storage.
	 *
	 *@param  cacheSize        Description of Parameter
	 *@exception  IOException  Description of Exception
	 */
	public synchronized void open(int cacheSize) throws IOException {
		if (!isOpen) {
			file = new CachedPersistentPageFile(treeName, cacheSize, false);
			// DEBUG System.out.println("CREANDO UN NUEVO PAGEFILE" + file.hashCode());
			if (file.tree != null) {
				throw new IllegalArgumentException("PageFile already in use by another rtree instance.");
			}
			// FIXME: Check for empty page file here.

			/*
			 *  if (file.treeType == -1) {
			 *  throw new IllegalArgumentException("PageFile is empty. Use some other RTree constructor.");
			 *  }
			 */
			file.tree = this;
			this.file = file;
			this.treeType = file.treeType;
			minimumLoad = Math.round(getNodeCapacity() * getFillFactor());
			if (minimumLoad < 2) {
				minimumLoad = 2;
			}
			isOpen = true;
		}
	}


	/**
	 *  Abre con valores por defecto
	 *
	 *@exception  IOException  Description of Exception
	 */
	public void open() throws IOException {
		open(128);
	}


	/**
	 *  Description of the Method
	 *
	 *@exception  IOException  Description of Exception
	 */
	public synchronized void close() throws IOException {
		if (isOpen) {
			this.stats.toString();
			file.close();
			//DEBUG System.out.println("CERRANDO EL PAGEFILE"+file.hashCode());
			isOpen = false;
		}
	}


	/**
	 *  Borra un Rect del RTree
	 *
	 *@param  r  Rect a borrar
	 *@return    Description of the Returned Value
	 */
	public long delete(com.vividsolutions.jts.geom.Envelope r) {
		if (r == null) {
			System.out.println("Sin Geometria");
			return NIL;
		}

//		HyperCube h = new HyperCube(new Point(r.xmin, r.ymin), new Point(r.xmax, r.ymax));
		HyperCube h = new HyperCube(new Point(r.getMinX(), r.getMinY()), new Point(r.getMaxX(), r.getMaxY()));
		return delete(h);
	}


	/**
	 *  Description of the Method
	 *
	 *@param  r    Description of Parameter
	 *@param  pos  Description of Parameter
	 */
	public void write(Envelope r, long pos) {
		if (r == null) {
			System.out.println("Sin Geometria");
			return;
		}

		AbstractNode root = file.readNode(0);
		HyperCube h = new HyperCube(new Point(r.getMinX(), r.getMinY()), new Point(r.getMaxX(), r.getMaxY()));
		Leaf l = root.chooseLeaf(h);
		l.insert(h, pos);
	}


	/**
	 *  Crea un iterador igual que los demas contendores.
	 *
	 *@param  r  Rectangulo sobre el que se iterara
	 *@return    iterador convencional
	 *@see       java.util.Iterator
	 */
	public RTreeIterator iterator(Envelope r) {
		//r.dump();
		return new RTreeIterator(r, file);
	}


	/**
	 *  Inserts a HyperCube into the tree, pointing to the data with the
	 *  specified object id.
	 *
	 *@param  h    The hypercube to insert.
	 *@param  oid  The object identifier of the data.
	 */
	public void insert(HyperCube h, long oid) {
		if (h == null) {
			throw new IllegalArgumentException("HyperCube cannot be null.");
		}
		if (h.getDimension() != file.dimension) {
			throw new IllegalArgumentException("HyperCube dimension different than RTree dimension.");
		}
		HyperCube ins = (HyperCube) h.clone();
		AbstractNode root = file.readNode(0);
		Leaf l = root.chooseLeaf(ins);
		l.insert(ins, oid);
	}


	/**
	 *  Deletes a HyperCube from the leaf level of the tree. If there is no leaf
	 *  containing a hypercube that matches the given hypercube, the tree is left
	 *  intact.
	 *
	 *@param  h  The HyperCube to delete.
	 *@return    The object id of the deleted entry, NIL if no matching entry was
	 *      found.
	 */
	public long delete(HyperCube h) {
		if (h == null) {
			throw new IllegalArgumentException("HyperCube cannot be null.");
		}
		if (h.getDimension() != file.dimension) {
			throw new IllegalArgumentException("HyperCube dimension different than RTree dimension.");
		}
		HyperCube del = (HyperCube) h.clone();
		AbstractNode root = file.readNode(0);
		Leaf l = root.findLeaf(del);
		return (l != null) ? l.delete(del) : NIL;
	}


	/**
	 *  Deletes a HyperCube from the leaf level of the tree. If there is no leaf
	 *  containing a hypercube that matches the given hypercube and the branch
	 *  'pos' , the tree is left intact.
	 *
	 *@param  r    Description of Parameter
	 *@param  pos  The Branch to delete.
	 *@return      The object id of the deleted entry, NIL if no matching entry
	 *      was found.
	 */
	public long delete(Envelope r, long pos) {
		if (r == null) {
//			Log.info("RTree.delete(null,"+pos+")Sin Geometria, no borramos de RTree");
			return NIL;
		}
		HyperCube del = new HyperCube(new Point(r.getMinX(), r.getMinY()), new Point(r.getMaxX(), r.getMaxY()));
		if (del == null) {
			throw new IllegalArgumentException("HyperCube cannot be null.");
		}
		if (del.getDimension() != file.dimension) {
			throw new IllegalArgumentException("HyperCube dimension different than RTree dimension.");
		}
		AbstractNode root = file.readNode(0);
		Leaf l = root.findLeaf(del, pos);
		return (l != null) ? l.delete(del, pos) : NIL;
	}


	// RRI
	/**
	 *  Description of the Method
	 *
	 *@param  r       Description of Parameter
	 *@param  oldPos  Description of Parameter
	 *@param  newPos  Description of Parameter
	 *@return         Description of the Returned Value
	 */
	public long update(Envelope r, long oldPos, long newPos) {
		AbstractNode root = file.readNode(0);
		HyperCube h = new HyperCube(new Point(r.getMinX(), r.getMinY()), new Point(r.getMaxX(), r.getMaxY()));
		return update(root, h, oldPos, newPos);
	}


	/**
	 *  Traverse all nodes in a breadth first manner. CAUTION: If the tree is not
	 *  memory resident, all nodes will be loaded into main memory.
	 *
	 *@param  v  The visitor pattern.
	 */
	public void traverseBreadthFirst(Visitor v) {
		if (v == null) {
			throw new IllegalArgumentException("Arguments cannot be null");
		}
		LinkedList l = new LinkedList();
		l.addLast(file.readNode(0));
		while (!l.isEmpty()) {
			AbstractNode n = (AbstractNode) l.removeFirst();
			v.visitNode(n);
			if (!n.isLeaf()) {
				for (int i = 0; i < n.usedSpace; i++) {
					l.addLast(((Index) n).getChild(i));
				}
			}
		}
	}


	/**
	 *  Traverse all nodes in a depth first manner. CAUTION: If the tree is not
	 *  memory resident, all nodes will be loaded into main memory.
	 *
	 *@param  v  The visitor pattern.
	 */
	public void traverseDepthFirst(Visitor v) {
		if (v == null) {
			throw new IllegalArgumentException("Arguments cannot be null");
		}
		Stack s = new Stack();
		s.push(file.readNode(0));
		while (!s.empty()) {
			AbstractNode n = (AbstractNode) s.pop();
			v.visitNode(n);
			if (!n.isLeaf()) {
				for (int i = n.usedSpace - 1; i >= 0; i--) {
					s.push(((Index) n).getChild(i));
				}
			}
		}
	}


	/**
	 *  Pre order traversal of tree nodes. CAUTION: If the tree is not memory
	 *  resident, all nodes will be loaded into main memory.
	 *
	 *@param  root  The sub-tree to start from.
	 *@param  v     The visitor pattern.
	 */
	public void traversePreOrder(AbstractNode root, Visitor v) {
		if (root == null || v == null) {
			throw new IllegalArgumentException("Arguments cannot be null.");
		}
		v.visitNode(root);
		if (root.isLeaf()) {
			for (int i = 0; i < root.usedSpace; i++) {
				v.visitData((HyperCube) root.data[i].clone(), root.branches[i]);
			}
		}
		else {
			for (int i = 0; i < root.usedSpace; i++) {
				traversePreOrder(((Index) root).getChild(i), v);
			}
		}
	}


	/**
	 *  Pre order traversal of tree nodes, starting from the root. CAUTION: If
	 *  the tree is not memory resident, all nodes will be loaded into main
	 *  memory.
	 *
	 *@param  v  The visitor pattern.
	 */
	public void traversePreOrder(Visitor v) {
		traversePreOrder(file.readNode(0), v);
	}


	/**
	 *  Post order traversal of tree nodes. CAUTION: If the tree is not memory
	 *  resident, all nodes will be loaded into main memory.
	 *
	 *@param  root  The sub-tree to start from .
	 *@param  v     The visitor pattern.
	 */
	public void traversePostOrder(AbstractNode root, Visitor v) {
		if (root == null || v == null) {
			throw new IllegalArgumentException("Arguments cannot be null.");
		}
		if (root.isLeaf()) {
			for (int i = 0; i < root.usedSpace; i++) {
				v.visitData((HyperCube) root.data[i].clone(), root.branches[i]);
			}
		}
		else {
			for (int i = 0; i < root.usedSpace; i++) {
				traversePostOrder(((Index) root).getChild(i), v);
			}
		}
		v.visitNode(root);
	}


	/**
	 *  Post order traversal of tree nodes, starting form the root. CAUTION: If
	 *  the tree is not memory resident, all nodes will be loaded into main
	 *  memory.
	 *
	 *@param  v  The visitor pattern.
	 */
	public void traversePostOrder(Visitor v) {
		traversePostOrder(file.readNode(0), v);
	}


	/**
	 *  Visit all data entries that intersect with the given HyperCube.
	 *
	 *@param  h  The HyperCube tested for intersection.
	 *@param  v  The visitor pattern.
	 */
	public void intersection(HyperCube h, Visitor v) {
		if (h == null || v == null) {
			throw new IllegalArgumentException("Arguments cannot be null.");
		}
		if (h.getDimension() != file.dimension) {
			throw new IllegalArgumentException("HyperCube dimension different than RTree dimension.");
		}
		Stack stack = new Stack();
		AbstractNode root = file.readNode(0);
		if (root.nodeMbb != null && root.nodeMbb.intersects(h)) {
			stack.push(root);
		}
		while (!stack.empty()) {
			AbstractNode n = (AbstractNode) stack.pop();
			v.visitNode(n);
			for (int i = 0; i < n.usedSpace; i++) {
				if (n.data[i].intersects(h)) {
					if (n.isLeaf()) {
						v.visitData((HyperCube) n.data[i].clone(), n.branches[i]);
					}
					else {
						stack.push(((Index) n).getChild(i));
					}
				}
			}
		}
	}


	// FIN ALEX

	/**
	 *  Visit all data entries that contain the given HyperCube.
	 *
	 *@param  h  The HyperCube tested for containment.
	 *@param  v  The visitor pattern.
	 */
	public void containment(HyperCube h, Visitor v) {
		if (h == null || v == null) {
			throw new IllegalArgumentException("Arguments cannot be null.");
		}
		if (h.getDimension() != file.dimension) {
			throw new IllegalArgumentException("HyperCube dimension different than RTree dimension.");
		}
		Stack stack = new Stack();
		AbstractNode root = file.readNode(0);
		if (root.nodeMbb != null && root.nodeMbb.contains(h)) {
			stack.push(root);
		}
		while (!stack.empty()) {
			AbstractNode n = (AbstractNode) stack.pop();
			v.visitNode(n);
			for (int i = 0; i < n.usedSpace; i++) {
				if (n.data[i].contains(h)) {
					if (n.isLeaf()) {
						v.visitData((HyperCube) n.data[i].clone(), n.branches[i]);
					}
					else {
						stack.push(((Index) n).getChild(i));
					}
				}
			}
		}
	}


	/**
	 *  Visit all data entries that contain the given point.
	 *
	 *@param  p  The point tested for containment.
	 *@param  v  The visitor pattern.
	 */
	public void containment(Point p, Visitor v) {
		containment(new HyperCube(p, p), v);
	}


	/**
	 *  Finds the nearest HyperCube to the given point. [King Lum Cheung and Ada
	 *  Wai-chee Fu: Enhanced Nearest Neighbor Search on the R-Tree]
	 *
	 *@param  p  The query point.
	 *@param  v  The visitor pattern.
	 */
	public void nearestNeighbor(Point p, Visitor v) {
		Data ret = nearestNeighborSearch(file.readNode(0), p, Float.POSITIVE_INFINITY, v);
		v.visitData(ret.data, ret.id);
	}


	/**
	 *  Validates minimum bounding boxes relationships.
	 */
	public void validateTree() {
		traverseDepthFirst(
					new Visitor() {
						public void visitNode(AbstractNode n) {
							if (n.data[0] != null) {
								HyperCube h = (HyperCube) n.data[0].clone();
								for (int j = 1; j < n.usedSpace; j++) {
									h = h.combinedMbb(n.data[j]);
								}
								if (!n.nodeMbb.equals(h)) {
									System.out.println("Combined node mbb error: " + n.pageNumber);
								}
							}
							AbstractNode p = n.getParent();
							if (p != null) {
								int j = -1;
								for (j = 0; j < p.usedSpace; j++) {
									if (p.branches[j] == n.pageNumber) {
										break;
									}
								}
								if (!p.data[j].equals(n.nodeMbb)) {
									System.out.println("Child-parent relationship error: " + n.pageNumber);
									System.out.println("p.data[" + j + "]=" + p.data[j] + "  n.nodeMbb=" + n.nodeMbb);
								}
							}
						}


						public void visitData(HyperCube h, long oid) {
						}
					});
	}


	/**
	 *  Print tree information and statistics.
	 *
	 *@return    A string representation of the object.
	 */
	public String toString() {
		String s;
		s = new String("Dimension: " + file.dimension + "\nFill factor: " + file.fillFactor + "\nCapacity: " + file.nodeCapacity + "\nMinimum load: " + minimumLoad + "\n" + stats.toString());
		return s;
	}


	/**
	 *  Overrides the finalize method to dispose of system resources or to
	 *  perform other cleanup when the RTree object is garbage collected.
	 *
	 *@exception  Throwable  Description of Exception
	 */
	protected void finalize() throws Throwable {
		close();
	}


	// ALEX

	/**
	 *  Returns a Vector with all nodes that intersect with the given HyperCube.
	 *  The nodes are returned in post order traversing
	 *
	 *@param  h     The given HyperCube that is tested for overlapping.
	 *@param  root  The node where the search should begin.
	 *@return       A Vector containing the appropriate nodes in the correct
	 *      order.
	 */
	ArrayList intersection(HyperCube h, AbstractNode root) {
		ArrayList v = new ArrayList();
		HyperCube cb = root.getNodeMbb();
		if (cb != null && cb.intersects(h)) {
			v.add(root);
			if (!root.isLeaf()) {
				for (int i = 0; i < root.usedSpace; i++) {
					if (root.data[i].intersects(h)) {
						ArrayList a = intersection(h, ((Index) root).getChild(i));
						v.addAll(a);
					}
				}
			}
		}
		return v;
	}


	/**
	 *  Returns an Enumeration with all nodes present in the tree that intersect
	 *  with the given HyperCube. The nodes are returned in post order traversing
	 *
	 *@param  h  The given HyperCube that is tested for overlapping.
	 *@return    An Enumeration containing the appropriate nodes in the correct
	 *      order.
	 */
	Enumeration intersection(HyperCube h) {


		class IntersectionEnum implements Enumeration {
			private boolean hasNext = true;
			private ArrayList nodes;
			private int index = 0;


			/**
			 *  Constructor for the IntersectionEnum object
			 *
			 *@param  hh  Description of Parameter
			 */
			public IntersectionEnum(HyperCube hh) {
				nodes = intersection(hh, file.readNode(0));
				if (nodes.isEmpty()) {
					hasNext = false;
				}
			}


			/**
			 *  Description of the Method
			 *
			 *@return    Description of the Returned Value
			 */
			public boolean hasMoreElements() {
				return hasNext;
			}


			/**
			 *  Description of the Method
			 *
			 *@return    Description of the Returned Value
			 */
			public Object nextElement() {
				if (!hasNext) {
					throw new NoSuchElementException("intersection");
				}
				Object c = nodes.get(index);
				index++;
				if (index == nodes.size()) {
					hasNext = false;
				}
				return c;
			}
		}
		;
		return new IntersectionEnum(h);
	}


	/**
	 *  Updates the HyperCube from the leaf level of the tree. If there is no
	 *  leaf containing a hypercube that matches the given hypercube and the
	 *  branch 'oldPos' , the tree is left intact.
	 *
	 *@param  r       the AbstractNode to update.
	 *@param  h       the HyperCube to update.
	 *@param  oldPos  the old value of the Branch.
	 *@param  newPos  the new value of the Branch.
	 *@return         Description of the Returned Value
	 */
	private long update(AbstractNode r, HyperCube h, long oldPos, long newPos) {
		long aux = NIL;
		for (int i = 0; (aux == NIL) && (i < r.usedSpace); i++) {
			if (r.isLeaf()) {
				if (r.branches[i] == oldPos) {
					r.branches[i] = newPos;
					file.writeNode(r);
					return 1;
				}
			}
			else {
				if (r.data[i].intersects(h)) {
					aux = update(file.readNode(r.branches[i]), h, oldPos, newPos);
				}
			}
		}
		return aux;
	}


	/**
	 *  Used for nearest neighbor searching in the RTree structure.
	 *
	 *@param  n        The node of the active branch list tested.
	 *@param  p        The query point.
	 *@param  nearest  The distance of <B>p</B> from current nearest hypercube
	 *      <B>h</B> .
	 *@param  v        The visitor pattern.
	 *@return          Description of the Returned Value
	 */
	private Data nearestNeighborSearch(AbstractNode n, Point p, float nearest, Visitor v) {
		Data ret = null;
		v.visitNode(n);
		if (n.isLeaf()) {
			for (int i = 0; i < n.usedSpace; i++) {
				float dist = n.data[i].getMinDist(p);
				if (dist < nearest) {
					ret = new Data(n.data[i], n.branches[i]);
					nearest = dist;
				}
			}
			return ret;
		}
		else {


			class ABLNode {
				long pageNumber;
				float minDist;


				/**
				 *  Constructor for the ABLNode object
				 *
				 *@param  pageNumber  Description of Parameter
				 *@param  minDist     Description of Parameter
				 */
				public ABLNode(long pageNumber, float minDist) {
					this.pageNumber = pageNumber;
					this.minDist = minDist;
				}
			}

			ABLNode[] abl = new ABLNode[n.usedSpace];
			for (int i = 0; i < n.usedSpace; i++) {
				abl[i] = new ABLNode(n.branches[i], n.data[i].getMinDist(p));
			}
			// sort ABL in ascending order of MINDIST from the query point.
			Arrays.sort(abl,
						new Comparator() {
							public int compare(Object o1, Object o2) {
								float f = ((ABLNode) o1).minDist - ((ABLNode) o2).minDist;
								// do not round the float here. It is wrong!
								return (f > 0) ? 1 : (f < 0) ? -1 : 0;
							}
						});
			// traverse all ABL nodes and prune irrelevant nodes according to the MINDIST heuristic.
			for (int i = 0; i < abl.length; i++) {
				if (abl[i].minDist <= nearest) {
					// recursively continue the search.
					Data d = nearestNeighborSearch(file.readNode(abl[i].pageNumber), p, nearest, v);
					// find the new nearest distance.
					if (d != null) {
						nearest = d.data.getMinDist(p);
						ret = d;
					}
				}
			}
			return ret;
		}
	}


	/**
	 *  Description of the Class
	 *
	 *@author    Chris Seguin
	 */
	class Data {
		/**
		 *  Description of the Field
		 */
		protected HyperCube data;
		/**
		 *  Description of the Field
		 */
		protected long id;


		/**
		 *  Constructor for the Data object
		 *
		 *@param  h   Description of Parameter
		 *@param  id  Description of Parameter
		 */
		public Data(HyperCube h, long id) {
			this.data = (HyperCube) h.clone();
			this.id = id;
		}
	}
}
// RTree
